/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/***************************************************************************
 *           KernelCom.C, S.Kesavan, kesavans@cs.uoregon.edu               *
 *           For the AWESIME version of pC++                               *
 *        								   *
 *           Provides definitions of all the functions                     *
 *           specified in runtime.h. These functions are                   *
 *           used by kernel.h and kernel-AWE.h                             *
 *                                           				   *
 ***************************************************************************/

#include <stdio.h>

#if defined(_KSR1_)
  #include "TreeBarTreeWakeup.h"
#else
  #include "Barrier.h"
#endif
#include "CpuMux.h"


#include "AweThread.h"            /* Thread class definition */
#include "se_event.h"             /* For tracing routines */

/* Some variable definitions from awelib.C   */
/* All these variables are global and shared */

#if defined(_KSR1_)
 extern TreeBarTreeWakeup AweGenBarr;
#else
 extern barrier AweGenBarr;	  /* general-purpose barrier */
#endif
extern void **AweElemAllocVector; /* vector of elem space pointers */

extern int Forked;		  /* 0:yet to fork, 1:forked */

/* Some function declarations from awelib.C */
extern unsigned long int GetClock();

/*************************************************************/
void Message(char *st, int flag)
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  if (flag == WARNING)
    {
      ThisThread->WarningCount++;
      printf("Warning %s\n",st);
      return;
    }
   if (flag == FATAL)
    {
      printf("Fatal error %s\n",st);
      /* there we may want to activate a general abort on all the processors??? */
      exit(1);
    }
}

void pcxx_Barrier()
{
  SE_EVENT (EC_RUNTIME, SE_BARRIER_ENTER,0);
  AweGenBarr.rendezvous();
  SE_EVENT (EC_RUNTIME, SE_BARRIER_EXIT,0);
}


void AweAllocElemSpace (int thread_id, int size)
{
  SE_EVENT (EC_RUNTIME, SE_ALLOC_ELEM_ENTER, thread_id);
  AweElemAllocVector[thread_id] = (void *) malloc (size);
  SE_EVENT (EC_RUNTIME, SE_ALLOC_ELEM_EXIT, (long) AweElemAllocVector[thread_id]);
}

void *AweGetElemSpacePtr (int thread_id)
{
  SE_EVENT (EC_RUNTIME, SE_GET_ELEM_PTR, thread_id);
  return (AweElemAllocVector[thread_id]);
}

void AweDealloc(void *ptr)
{
  SE_EVENT (EC_RUNTIME, SE_DEALLOC_ENTER, (long) ptr);
  free(ptr);
  SE_EVENT (EC_RUNTIME, SE_DEALLOC_EXIT, 0);
}

int pcxx_TotalNodes()
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  return ThisThread->NumThreads;
}

int pcxx_NodeNumber()
{
  AweThread *ThisThread;

  /* No threads, sequential part */
  if (!Forked)
     return 0;

  /* Threads forked, but no threads in the multiplexor, */
  /* probably because they are all done */
  if ( (CpuMux::CurrentThread()) == NULL)
     return 0;

  /* Else return the identity of the thread that is */
  /* currently running in the multiplexor */
  ThisThread = (AweThread *)(CpuMux::CurrentThread());
  return ThisThread->MyId;
}

int GetNumberOfCollection()
{ 
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  return ThisThread->NumberOfCollection;
}

void IncNumberOfCollection()
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  if ( (ThisThread->NumberOfCollection) >= MaxCollection)
    Message("Too many collections.",FATAL);
  ThisThread->NumberOfCollection++;
}

void SetTableOfCollection (int num, char *addr)
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  ThisThread->TableOfCollection[num] = (PTR_INFO) addr;

}

void SetNextCollectionTraversalLevel (char *addr)
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  ThisThread->TraversalLevel++;
  if ( (ThisThread->TraversalLevel) >= MAXPARAL)
    Message("ResetLocal(): Too much parallelism.",FATAL);
  ThisThread->CurrentCollectionTraversal[ (ThisThread->TraversalLevel) ] = addr;
}

char *GetCollectionTraversalLevel ()
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  return ThisThread->CurrentCollectionTraversal[ (ThisThread->TraversalLevel) ];
}

void DecCollectionTraversalLevel ()
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  ThisThread->TraversalLevel--;
}

void ResetCollectionTraversalLevel ()
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  ThisThread->TraversalLevel = -1;
}

void pcxx_Poll()
{
  return;
}

/**************************************************************************
 ********************** Kernel Requests ***********************************
 **************************************************************************/
/* These functions are used only in distributed memory versions   */
/* and Uniproc. For shared memory systems and awesime these are   */
/* just disabled by having them print an error message            */

int IamtheManager(int id)
{
  Message("IamtheManager called; This is fatal Error ",FATAL);
  return 1;
}

int WhereIsData(int id, int info)
{
  Message("WhereIsData called; This is fatal Error ",FATAL);
  return 0;
}

int AskForData(int where, int index, int offset, int info)
{
  Message("AskForData called; This is fatal Error ",FATAL);
  return 0;
}

int AskForDataBlock(int where, int index, int info, int *buffer)
{
  Message("AskForData called; This is fatal Error", FATAL);
  return 0;
}


/************************************************************************
 ************************** Timer functions *****************************
 ************************************************************************/

int pcxx_UserTimerClear(int i)
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());

  if ((i<0) || (i>=64)) Message("Too many timers",FATAL);
  SE_EVENT (EC_TIMER, SE_TIMER_CLEAR, i);
  ThisThread->TimerStart[i] = 0;
  ThisThread->TimerStop[i] = 0;
  ThisThread->TimerElapsed[i]  = 0;
  return 1;
}

int pcxx_UserTimerStart(int i)
{ 
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());
  if ((i<0) || (i>=64)) Message("Too many timers",FATAL);
  SE_EVENT (EC_TIMER, SE_TIMER_START + i, 0);
  ThisThread->TimerStart[i] =  ((double) GetClock()) / 1000000.0;
  return 1;
}

int pcxx_UserTimerStop(int i)
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());
  if ((i<0) || (i>=64)) Message("Too many timers",FATAL);
  SE_EVENT (EC_TIMER, SE_TIMER_STOP + i, 0);
  ThisThread->TimerStop[i] = (((double) GetClock()) / 1000000.0)
                             - ThisThread->TimerStart[i];
  ThisThread->TimerElapsed[i] += ThisThread->TimerStop[i];
  return 1;
}

double pcxx_UserTimerElapsed(int i)
{
  AweThread *ThisThread;

  ThisThread = (AweThread *)(CpuMux::CurrentThread());
  if ((i<0) || (i>=64)) Message("Too many timers",FATAL);
  return ThisThread->TimerElapsed[i];
}

/**************************************************************************
 ********************** General Routines **********************************
 **************************************************************************/

/*************************************************************/
/* Simple bit shuffle for quadtree numbering to 2d grid (PHB) */
unsigned long BitShuffle(unsigned long p, unsigned long q)
{
unsigned long x,y,z;
int i,j,bit1,bit2;
 
/* Mask off high 16 bits, they are out of range... */
x = (1<<16)-1 & p;
y = (1<<16)-1 & q;
 
if ((x != p) || (y != q))
     Message("Error, BitShuffle() parm too big, must be <= 16 bits\n",FATAL);
 
/* Shuffle-d-bits */
z = 0;
for (i=15; i>=0; i--)
  {
  bit1 = (x >> i) & 1;
  bit2 = (y >> i) & 1;
  z = (z << 2) | (bit1 << 1) | (bit2 << 0);
  }
 
return z;
}  
  
 
/*************************************************************/
/* Simple bit shuffle for quadtree numbering to 2d grid (PHB) */
void BitUnshuffle(unsigned long *p, unsigned long *q, unsigned long x)
{
int i,bit1,bit2;
 
*p = 0;
*q = 0;
for (i=31; i>=1; i=i-2)
  { 
  bit1 = (x >> i)     & 1;
  bit2 = (x >> (i-1)) & 1;
  *p = (*p << 1) | bit1;
  *q = (*q << 1) | bit2;
}
}
 
 
/*************************************************************/
/* Maps (i,j) elements in grid to linear number (x) by blocks (PHB) */
int QGridToProc(unsigned i, unsigned j, int M)
{
unsigned long jprime, block;
int procnum;
 
jprime = j % (unsigned long) M;
block = j / (unsigned long) M;
 
procnum = (int) ((unsigned long) M * (unsigned long) M * block + 
                     BitShuffle(i,jprime));
return procnum;
}
