/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/****************************************************************/
/* awelib : Awe run-time support library			*/
/****************************************************************/

#include <stdio.h>
#include <sys/time.h>

#if defined(_KSR1_)
# include "TreeBarTreeWakeup.h"
#else
# include "Barrier.h"      /* Awesime Barrier class */
#endif
#include "Thread.h"       /* Awesime thread class */
#include "CpuMux.h"       /* Awesime CpuMux class */
#include "MultiFifoMux.h" /* Awesime Mux class */

#include "AweThread.h"    /* runtime system's thread class */              
#include "se_event.h"	  /* For tracing routines */

#if defined(_KSR1_)
  TreeBarTreeWakeup AweGenBarr;
#else
  barrier AweGenBarr(0);	  /* general-purpose barrier */
#endif
pcxx_PARAM_t pcxx_param;    /* structure containing all pC++ parameters */
long se_secs;             /* wallclock time at the beginning */
void **AweElemAllocVector;/* vector of elem space pointers */
int Forked=0;		  /* 0:yet to fork, 1:forked */



/**********************************
 *     Timer support functions    *
 **********************************/

unsigned long int SetClock()
{
  struct timeval tp;
  gettimeofday (&tp,NULL);
  se_secs = tp.tv_sec;
  return 0L;
}

unsigned long int GetClock()
{
  struct timeval tp;
  gettimeofday(&tp,NULL);
  return ( (tp.tv_sec - se_secs) * 1000000 + tp.tv_usec);
}

/************************************
 * AweSetup() : Initial setup       *
 ************************************/

# define GetNextArg(variable,str) \
        if ( (i+1) == *argc )   \
          printf ("%s: missing argument for %s - ignored.\n", argv[0], str); \
        else                    \
        {                       \
          found = 1;            \
          variable = argv[i+1]; \
        }

int InitParam (int *argc, char **argv)
{
  char *procstr   = NULL;
  char *eventstr  = NULL;
  char *filestr   = NULL;
  char *threadstr = NULL;
  char *shmemsizestr = NULL;
  char *stacksizestr = NULL;
  char *end;
  int i, j, found;

  for (i=1; i<*argc; i++)
  {
    if ( strncmp (argv[i], "-pcxx_", 6) == 0 )
    {
      found = 0;

      if ( strcmp (argv[i], "-pcxx_NUMPROC") == 0 )
	{ GetNextArg (procstr, "pcxx_NUMPROC"); }
      else if ( strcmp (argv[i], "-pcxx_NUMTHREAD") == 0 )
        { GetNextArg (threadstr, "pcxx_NUMTHREAD"); }
      else if ( strcmp (argv[i], "-pcxx_EVENTCLASS") == 0 )
        { 
	  GetNextArg (eventstr, "pcxx_EVENTCLASS");
# ifndef TRACING_ON
          printf ("warning: no tracing - event class ignored.\n");
# endif
          }
      else if ( strcmp (argv[i], "-pcxx_TRACEFILE") == 0 )
        { 
	  GetNextArg (filestr, "pcxx_TRACEFILE");
# ifndef TRACING_ON
          printf ("warning: no tracing - trace filename ignored.\n");
# endif
         }
      else if (strcmp (argv[i], "-pcxx_SHMEMSIZE") == 0 )
        { GetNextArg (shmemsizestr, "pcxx_SHMEMSIZE"); }
      else if (strcmp (argv[i], "-pcxx_STACKSIZE") == 0 )
        { GetNextArg (stacksizestr, "pcxx_STACKSIZE"); }

      if ( found )
      {
        *argc -= 2;
        for (j=i; j<*argc; j++) argv[j] = argv[j+2];
        i--;
      }
    }
  }

  if ( (procstr == NULL) && ((procstr = getenv ("pcxx_NUMPROC")) == NULL) )
    pcxx_param.numproc = DEF_NUMPROC;
  else if ( (pcxx_param.numproc = atoi(procstr)) <= 0 )
  {
    printf ("%s: invalid number of procs %s - ignored.\n", argv[0], procstr);
    pcxx_param.numproc = DEF_NUMPROC;
  }

  if ( (threadstr == NULL) && ((threadstr = getenv ("pcxx_NUMTHREAD")) == NULL) )
    pcxx_param.numthread = DEF_NUMTHREAD;
  else if ( (pcxx_param.numthread = atoi(threadstr)) <= 0 )
  {
    printf ("%s: invalid number of threads %s - ignored.\n", argv[0],threadstr);
    pcxx_param.numthread = DEF_NUMTHREAD;
  }

  if ( (shmemsizestr == NULL) && ((shmemsizestr = getenv ("pcxx_SHMEMSIZE")) == NULL) )
    pcxx_param.shmemsize = DEF_SHMEMSIZE;
  else if ( (pcxx_param.shmemsize = atoi(shmemsizestr)) <= 0 )
  {
    printf ("%s: invalid shmemsize %s - ignored.\n", argv[0],shmemsizestr);
    pcxx_param.shmemsize = DEF_SHMEMSIZE;
  }

  if ( (stacksizestr == NULL) && ((stacksizestr = getenv ("pcxx_STACKSIZE")) == NULL) )
    pcxx_param.stacksize = DEF_STACKSIZE;
  else if ( (pcxx_param.stacksize = atoi(stacksizestr)) <= 0 )
  {
    printf ("%s: invalid stacksize %s - ignored.\n", argv[0],stacksizestr);
    pcxx_param.stacksize = DEF_STACKSIZE;
  }

# ifdef TRACING_ON
  if ( (filestr == NULL) && ((filestr = getenv ("pcxx_TRACEFILE")) == NULL) )
    pcxx_param.tracefile = DEF_TRACEFILE;
  else
    pcxx_param.tracefile = filestr;

  if ( eventstr == NULL ) eventstr = getenv ("pcxx_EVENTCLASS");
  se_set_and_parse_eventclass (argv[0], eventstr);
# endif
  return 0;
}

int AweSetup()		
{
  SE_EVENT (EC_BASIC, SE_SETUP_ENTER, pcxx_param.numthread);

  SetClock();

  DuplicateGlobalVars(pcxx_param.numthread);

  /* allocate AweElemAllocVector */
  AweElemAllocVector = (void **) malloc (pcxx_param.numthread * sizeof(void *));

  /* initialize barriers */
#if defined(_KSR1_)
  AweGenBarr.initbarrier(pcxx_param.numthread, pcxx_param.numproc);
#else
  AweGenBarr.height(pcxx_param.numthread);
#endif

  SE_EVENT (EC_BASIC, SE_SETUP_EXIT, pcxx_param.numthread);

  return(0);
}

/************************************************************************
 * AweFork : start parallel processing					*
 *  argc, argv : passed down to each thread				*
 ************************************************************************/

int AweFork(int argc, char **argv)
{
  int i;
  MultiFifoMux Cpu;

  SE_EVENT (EC_BASIC, SE_FORK_ENTER, 0);

  /* fork */
  for(i = 0; i < pcxx_param.numthread; i++){
    CpuMux::add(new AweThread(pcxx_param.numthread, i, pcxx_param.stacksize, argc, argv));
    }

  Forked = 1;
  Cpu.fireItUp(pcxx_param.numproc, pcxx_param.shmemsize);
  Forked = 0;

  SE_EVENT (EC_BASIC, SE_FORK_EXIT, 0);

  return(0);
}

main(int argc,char **argv)
{ 
  InitParam(&argc, argv);
  se_init(pcxx_param.tracefile); 
  AweSetup();
  AweFork(argc, argv);
  se_close(); 
}

/*****************************************************************************
 * Constructor for AweThread      					     *
 *  num_threads : number of threads in the system			     *
 *  id : thread's id assigned by AweFork 				     *
 *  stacksize : thread's stacksize					     *
 *  argc, argv : will be passed to Processor_Main, in later implementations  *
 *****************************************************************************/

AweThread::AweThread(int num_threads, int id, int stacksize, int argc, char **argv)
: Thread ("AweThread", stacksize)
{
  NumThreads = num_threads;
  MyId = id;
  Argc = argc;
  Argv = argv;
  NumberOfCollection = 0;
  WarningCount = 0;
  TraversalLevel = -1;
}

/*********************************
 *  main procedure of AweThread  *
 *********************************/

void
AweThread::main()
{
#if defined(_KSR1_)
  AweGenBarr.checkIn();
#endif
  SE_EVENT(EC_BASIC, SE_PROC_MAIN_CALL, 0);
  Processor_Main();
  SE_EVENT(EC_BASIC, SE_PROC_MAIN_RETURN, 0);
#if defined(_KSR1_)
  AweGenBarr.checkOut();
#endif
}
